ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     setting input parameters                                    c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine parameter
      implicit none
c input parameters
      real*8 pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,mw,mw2
      real*8 gw,mz,mz2,gz,mh,mh2,mxe,mxe2,ml(3),ml2(3),mqp(3)
      real*8 mqp2(3),mqm(3),mqm2(3)
      common/param/pi,el,alpha,alpha0,alphaz,alphas,GF,cw,cw2,sw,sw2,
     *  mw,mw2,gw,mz,mz2,gz,mh,mh2,mxe,mxe2,ml,ml2,mqp,mqp2,mqm,mqm2
      alpha0  = 1d0/137.0359895d0     ! alpha(0)
      alphaz  = 1d0/128.887d0         ! alpha(mz)
      alphas  = 0.119d0               ! alpha_s
      GF      = 1.16637D-5            ! fermi constant
      mw      = 80.35d0               ! w mass
      mz      = 91.1867d0             ! z mass
      mh      = 150d0                 ! higgs mass
      gw      = 2.05d0                ! w width
      gz      = 2.49471d0             ! z width
      ml(1)   = 0.51099907d-3         ! electron
      ml(2)   = 0.105658389d0         ! muon
      ml(3)   = 1.77705d0             ! tau
      mqm(1)  = 0.00485d0             ! d-quark
      mqp(1)  = 0.00485d0             ! u-quark
      mqm(2)  = 0.150d0               ! s-quark
      mqp(2)  = 1.55d0                ! c-quark
      mqm(3)  = 4.5d0                 ! b-quark
      mqp(3)  = 174.17d0              ! t-quark
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     photon recombination subroutine                             c
c                                                                 c
c input:                                                          c 
c ------                                                          c
c four-momenta of external particles: pin(7,0:3)                  c 
c pbeam(1,0:3),pbeam(2,0:3): beams                                c
c pin(3-6,0:3): final-state fermions                              c
c pin(7,0:3): photon                                              c
c pin(n,0): energy, pin(n,1-3): x,y,z component                   c
c                                                                 c
c output:                                                         c
c -------                                                         c
c pout(7,0:3): four-momenta after eventual recombination          c
c nocut: 1: event passes cuts, 0: event is rejected               c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine recombination(pbeam,pin,pout,nocut)
      implicit none
c local variables
      real*8 pbeam(2,0:3),pin(7,0:3),pout(7,0:3),k(0:3),s(6)
      real*8 smin,cthetamax,pvec(7),pbeamvec(2),ctheta(7),pi
      integer i1,i2,i3,fcomb,nocut,comb
c cuts
      real*8 ecut(7),scut(7,7),ccut(7,7),precomb(4),precombg(2)
      integer scuts,srecomb,srecombg
c processpar
      real*8 mv(3),mv2(3),gv(3),qcdfac,qcdfac2,q(6)
      complex*16 cvf(-1:1,3,6),cvw(2)
      integer qprop,quark(6),iso(6),gen(6),dir(7),pol(7),color
      character*6 fname(6),dfname(6)
      common/cuts/ecut,scut,ccut,precomb,precombg,scuts,
     *  srecomb,srecombg
      common/processpar/cvf,cvw,mv,mv2,gv,qcdfac,qcdfac2,
     *  q,qprop,quark,iso,gen,dir,pol,color,fname,dfname
c initialization
      pi=4d0*datan(1d0)
      nocut=0
      fcomb=0
      comb=0
      do i1=1,7
      do i2=0,3
        pout(i1,i2)=pin(i1,i2)
      enddo
      enddo
c srecomb=1: TH, default as defined in LEP2MC report
c srecomb=3: TH without separation cuts
      if(srecomb.eq.1.or.srecomb.eq.3)then
c separation cut on angle between photon and beams:
         pvec(7)=dsqrt(pin(7,1)**2+pin(7,2)**2+pin(7,3)**2)
         do i1=1,2
           pbeamvec(i1)=dsqrt(pbeam(i1,1)**2+pbeam(i1,2)**2
     *                        +pbeam(i1,3)**2)
           ctheta(i1)=(pin(7,1)*pbeam(i1,1)+pin(7,2)*pbeam(i1,2)
     *       +pin(7,3)*pbeam(i1,3))/pvec(7)/pbeamvec(i1)
           if(ctheta(i1).gt.dcos(precomb(1)*pi/180d0))then
             do i2=0,3
               pout(7,i2)=0d0
             enddo
             call cut(pbeam,pout,4,nocut)
             return
           endif
         enddo
c determine minimal invariant mass smin(fcomb,photon)
c and recombine photon with p(fcomb) within the recombination cut
         i3=3
         if(srecomb.eq.3)i3=1
         do i1=i3,6
           if(q(i1).ne.0d0)then
             do i2=0,3
               k(i2)=pin(i1,i2)+pin(7,i2)
             enddo
             s(i1)=k(0)*k(0)-k(1)*k(1)-k(2)*k(2)-k(3)*k(3)
             if(s(i1).lt.precomb(3)**2.or.pin(7,0).lt.precomb(2))then
               comb=1
             endif
           endif
         enddo
         smin=1d10
         do i1=i3,6
           if(q(i1).ne.0d0.and.smin.gt.s(i1))then  
             smin=s(i1)
             fcomb=i1
           endif
        enddo
c srecomb=2: EXP, as defined in LEP2MC report
      elseif(srecomb.eq.2)then
c separation cut on angle between photon and beams:
         pvec(7)=dsqrt(pin(7,1)**2+pin(7,2)**2+pin(7,3)**2)
         do i1=1,2
           pbeamvec(i1)=dsqrt(pbeam(i1,1)**2+pbeam(i1,2)**2
     *                        +pbeam(i1,3)**2)
           ctheta(i1)=(pin(7,1)*pbeam(i1,1)+pin(7,2)*pbeam(i1,2)
     *       +pin(7,3)*pbeam(i1,3))/pvec(7)/pbeamvec(i1)
           if(ctheta(i1).gt.dcos(precomb(1)*pi/180d0))then
             do i2=0,3
               pout(7,i2)=0d0
             enddo
             call cut(pbeam,pout,4,nocut)
             return
           endif
         enddo 
c determine maximal angle cthetamax(fcomb,photon)
c and recombine photon with p(fcomb) within the recombination cut
         pvec(7)=dsqrt(pin(7,1)**2+pin(7,2)**2+pin(7,3)**2)
         do i1=3,6
           if(q(i1).ne.0d0)then
             pvec(i1)=dsqrt(pin(i1,1)**2+pin(i1,2)**2+pin(i1,3)**2)
             ctheta(i1)=(pin(i1,1)*pin(7,1)+pin(i1,2)*pin(7,2)
     *         +pin(i1,3)*pin(7,3))/pvec(i1)/pvec(7)
             if((quark(i1).eq.0.and.
     *         ctheta(i1).gt.dcos(precomb(3)*pi/180d0)).or.
     *         (quark(i1).eq.1.and.
     *         ctheta(i1).gt.dcos(precomb(4)*pi/180d0)).or.
     *         pin(7,0).lt.precomb(2))then
               comb=1
             endif
           endif
        enddo
        cthetamax=-1d0
        do i1=3,6
           if(q(i1).ne.0d0.and.cthetamax.lt.ctheta(i1))then  
             cthetamax=ctheta(i1)
             fcomb=i1
           endif
        enddo
      endif
c photon is recombined with fermion fcomb if comb=1
      if(comb.eq.1)then
        do i1=0,3
          if(fcomb.gt.2) 
     $          pout(fcomb,i1)=pin(fcomb,i1)+pin(7,i1)
          pout(7,i1)=0d0
        enddo
        call cut(pbeam,pout,4,nocut)
        return
      endif
c separation cuts
      call cut(pbeam,pout,5,nocut)
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     gluon recombination subroutine                              c
c                                                                 c
c input:                                                          c 
c ------                                                          c
c four-momenta of external particles: pin(7,0:3)                  c 
c pbeam(1,0:3),pbeam(2,0:3): beams                                c
c pin(3-6,0:3): final-state fermions                              c
c pin(7,0:3): gluon                                               c
c pin(n,0): energy, pin(n,1-3): x,y,z component                   c
c                                                                 c
c output:                                                         c
c -------                                                         c
c pout(7,0:3): four-momenta after eventual recombination          c
c nocut: 1: event passes cuts, 0: event is rejected               c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine grecombination(pbeam,pin,pout,nocut)
      implicit none
c local variables
      real*8 pbeam(2,0:3),pin(7,0:3),pout(7,0:3),k(0:3),s(6)
      real*8 smin,cthetamax,pvec(7),ctheta(7),pi
      integer i1,i2,fcomb,nocut,comb
c cuts
      real*8 ecut(7),scut(7,7),ccut(7,7),precomb(4),precombg(2)
      integer scuts,srecomb,srecombg
c processpar
      real*8 mv(3),mv2(3),gv(3),qcdfac,qcdfac2,q(6)
      complex*16 cvf(-1:1,3,6),cvw(2)
      integer qprop,quark(6),iso(6),gen(6),dir(7),pol(7),color
      character*6 fname(6),dfname(6)
      common/cuts/ecut,scut,ccut,precomb,precombg,scuts,
     *  srecomb,srecombg
      common/processpar/cvf,cvw,mv,mv2,gv,qcdfac,qcdfac2,
     *  q,qprop,quark,iso,gen,dir,pol,color,fname,dfname
c initialization
      pi=4d0*datan(1d0)
      nocut=0
      fcomb=0
      comb=0
      do i1=1,7
      do i2=0,3
        pout(i1,i2)=pin(i1,i2)
      enddo
      enddo
c srecombg=1: TH, default
      if(srecombg.eq.1)then
c determine minimal invariant mass smin(fcomb,gluon)
c and recombine gluon with p(fcomb) within the recombination cut
         do i1=3,6
           if(q(i1).ne.0d0.and.q(i1).ne.-1d0)then
             do i2=0,3
               k(i2)=pin(i1,i2)+pin(7,i2)
             enddo
             s(i1)=k(0)*k(0)-k(1)*k(1)-k(2)*k(2)-k(3)*k(3)
             if(s(i1).lt.precombg(2)**2.or.pin(7,0).lt.precombg(1))then
               comb=1
             endif
           endif
         enddo
         smin=1d10
         do i1=3,6
           if(q(i1).ne.0d0.and.q(i1).ne.-1d0.and.smin.gt.s(i1))then  
             smin=s(i1)
             fcomb=i1
           endif
        enddo
c srecombg=2: EXP
      elseif(srecombg.eq.2)then
c determine maximal angle cthetamax(fcomb,gluon)
c and recombine gluon with p(fcomb) within the recombination cut
         pvec(7)=dsqrt(pin(7,1)**2+pin(7,2)**2+pin(7,3)**2)
         do i1=3,6
           if(q(i1).ne.0d0.and.q(i1).ne.-1d0)then
             pvec(i1)=dsqrt(pin(i1,1)**2+pin(i1,2)**2+pin(i1,3)**2)
             ctheta(i1)=(pin(i1,1)*pin(7,1)+pin(i1,2)*pin(7,2)
     *         +pin(i1,3)*pin(7,3))/pvec(i1)/pvec(7)
             if(ctheta(i1).gt.dcos(precombg(2)*pi/180d0).or.
     *         pin(7,0).lt.precombg(1))then
               comb=1
             endif
           endif
        enddo
        cthetamax=-1d0
        do i1=3,6
           if(q(i1).ne.0d0.and.q(i1).ne.-1d0.and.
     *       cthetamax.lt.ctheta(i1))then  
             cthetamax=ctheta(i1)
             fcomb=i1
           endif
        enddo
      endif
c gluon is recombined with quark fcomb if comb=1
      if(comb.eq.1)then
        do i1=0,3
          pout(fcomb,i1)=pin(fcomb,i1)+pin(7,i1)
          pout(7,i1)=0d0
        enddo
        call cut(pbeam,pout,4,nocut)
        return
      endif
c separation cuts 
c separation cuts on the gluon are the same as for the photon !
      call cut(pbeam,pout,5,nocut)
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     cut subroutine                                              c
c                                                                 c
c input:                                                          c 
c ------                                                          c
c four-momenta of external particles: p(7,0:3)                    c 
c p(1,0:3),p(2,0:3): beams, p(3-6,0:3): final-state fermions      c
c p(7,0:3): photon                                                c
c p(n,0): energy, p(n,1-3): x,y,z component                       c
c nparticle: number of particles in the final state               c
c                                                                 c
c Note that p(1,0:3) and p(2,0:3) are the momenta after ISR !     c
c pbeam(2,0:3) are the beam momenta as defined in racoonww.f!     c
c                                                                 c
c output:                                                         c
c -------                                                         c
c nocut: 1: event passes cuts, 0: event is rejected               c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine cut(pbeam,p,nparticle,nocut)
      implicit none
c local variables
      real*8 pbeam(2,0:3),p(7,0:3),q(0:3),pvec(7),pbeamvec(2),s,prod
      integer i1,i2,i3,nparticle,nocut
c cuts
      real*8 ecut(7),scut(7,7),ccut(7,7),precomb(4),precombg(2)
      integer scuts,srecomb,srecombg
      common/cuts/ecut,scut,ccut,precomb,precombg,scuts,
     *  srecomb,srecombg
      nocut=0
c apply energy cuts
      do i1=3,nparticle+2
        if(p(i1,0).lt.ecut(i1))return
      enddo
c apply invariant-mass cuts
      do i1=3,nparticle+2
      do i2=i1+1,nparticle+2
        do i3=0,3
          q(i3)=p(i1,i3)+p(i2,i3)
        enddo
        s=q(0)*q(0)-q(1)*q(1)-q(2)*q(2)-q(3)*q(3)
        if(s.lt.scut(i1,i2))return
      enddo
      enddo
c apply angular cuts
      do i1=1,2
        pbeamvec(i1)=dsqrt(pbeam(i1,1)**2+pbeam(i1,2)**2
     *                     +pbeam(i1,3)**2)
      enddo
      do i1=3,nparticle+2
        pvec(i1)=dsqrt(p(i1,1)**2+p(i1,2)**2+p(i1,3)**2)
      enddo
      do i1=1,2
      do i2=3,nparticle+2
        prod=(pbeam(i1,1)*p(i2,1)+pbeam(i1,2)*p(i2,2)
     *        +pbeam(i1,3)*p(i2,3))
        if(prod.gt.ccut(i1,i2)*pbeamvec(i1)*pvec(i2))return
      enddo
      enddo
      do i1=3,nparticle+2
      do i2=i1+1,nparticle+2
        prod=(p(i1,1)*p(i2,1)+p(i1,2)*p(i2,2)+p(i1,3)*p(i2,3))
        if(prod.gt.ccut(i1,i2)*pvec(i1)*pvec(i2))return
      enddo
      enddo
c non-standard cuts can be inserted here
c
      nocut=1
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     subroutine settings                                         c
c                                                                 c
c definition of histograms is done here !                         c
c                                                                 c
c input:                                                          c
c ------                                                          c
c p(7,0:3) : four-momenta of external particles                   c
c smc=2,nw=6: p(7,0:3) photon momentum                            c
c       nw=7: p(7,0:3) gluon momentum                             c
c smc=1 or 3,nw=7: p(7,0:3) photon momentum                       c
c            nw=8: p(7,0:3) gluon momentum                        c
c y        : weight(nw) of event n                                c
c energy   : CMF energy                                           c
c q(6)     : charges of external particles                        c
c n        : number of event                                      c
c steps    : flag for subroutine histogram                        c
c            steps=1: initialization                              c
c            steps=2: start filling histograms                    c
c            steps=3: finish filling of histograms                c
c            steps=4: print data for histograms                   c
c                                                                 c
c Note that p(1,0:3) and p(2,0:3) are the momenta after ISR !     c
c pbeam(2,0:3) are the beam momenta as defined in racoonww.f!     c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine settings(nw,pbeam,p,y,energy,n,steps)
      implicit none
c local variables
      integer maxe
      parameter(maxe=29)
      real*8 energy,pbeam(2,0:3),p(7,0:3),s(7,7),y,x,pi,denum
      real*8 kp(0:3),km(0:3),kpvec,kmvec,pvec(7),pbeamvec(2)
      real*8 q1(0:3),q2(0:3),q3(0:3),yphoton,q1vec,q2vec
      integer i1,i2,i3,n,steps,nw
      character*10 name
c processpar
      real*8 mv(3),mv2(3),gv(3),qcdfac,qcdfac2,q(6)
      complex*16 cvf(-1:1,3,6),cvw(2)
      integer qprop,quark(6),iso(6),gen(6),dir(7),pol(7),color
      character*6 fname(6),dfname(6)
c options
      integer smc,sborn4,sborn5,sborng5,sisr,src,son,sqcdepem
      integer ssigepem4,ssigepem5,ssigepemg5,scoultree,sunw,spaw,satgc
c rcoptions
      integer qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
      common/processpar/cvf,cvw,mv,mv2,gv,qcdfac,qcdfac2,
     *  q,qprop,quark,iso,gen,dir,pol,color,fname,dfname
      common/options/smc,sborn4,sborn5,sborng5,sisr,src,son,sqcdepem,
     *  ssigepem4,ssigepem5,ssigepemg5,scoultree,sunw,spaw,satgc
      common/rcoptions/qborn,qreal,qqcd,qbos,qferm,qsoft,qnf,qfast
      pi=4d0*datan(1d0)
      if(y.eq.0d0.and.steps.eq.2)return
c      call qgc(pbeam,p,y,6) ! QGC, neutrino momentum, additional cut
c calculation of sum of pairs of momenta and the
c according invariant masses squared:
      do i1=3,7
      do i2=i1+1,7
        do i3=0,3
          q1(i3)=p(i1,i3)+p(i2,i3)
        enddo
        s(i1,i2)=q1(0)*q1(0)-q1(1)*q1(1)-q1(2)*q1(2)-q1(3)*q1(3)
        s(i2,i1)=s(i1,i2)
      enddo
      enddo
c calculation of the W+ and W- momenta:
c (not correct for mixed processes (i.e. kp(km)
c do not correspond to the W+(W-) momentum but to the
c gamma/Z momentum) !)
      do i1=0,3
        kp(i1)=p(3,i1)+p(4,i1)
        km(i1)=p(5,i1)+p(6,i1)
      enddo
      kpvec=dsqrt(kp(1)*kp(1)+kp(2)*kp(2)+kp(3)*kp(3))
      kmvec=dsqrt(km(1)*km(1)+km(2)*km(2)+km(3)*km(3))
      do i1=1,2
        pbeamvec(i1)=dsqrt(pbeam(i1,1)**2+pbeam(i1,2)**2+pbeam(i1,3)**2)
      enddo
      do i1=3,7
        pvec(i1)=dsqrt(p(i1,1)**2+p(i1,2)**2+p(i1,3)**2)
      enddo
c total cross section
      name='dat.01'
      call histogram(-0.5d0,0.5d0,0d0,y,1,n,name,1,steps)      
c invariant mass distributions
      name='dat.02'
      call histogram(75d0,85d0,dsqrt(s(3,4)),y,2,n,name,50,steps) 
      name='dat.03'
      call histogram(75d0,85d0,dsqrt(s(5,6)),y,3,n,name,50,steps)
c production angle distributions
      x=2d0
      if(kpvec*pbeamvec(1).gt.0d0)then
        x=(kp(1)*pbeam(1,1)+kp(2)*pbeam(1,2)+kp(3)*pbeam(1,3))
     *    /kpvec/pbeamvec(1)
      endif
      name='dat.04'
      call histogram(-1d0,1d0,x,y,4,n,name,50,steps)
      x=2d0
      if(kmvec*pbeamvec(2).gt.0d0)then
        x=(km(1)*pbeam(2,1)+km(2)*pbeam(2,2)+km(3)*pbeam(2,3))
     *    /kmvec/pbeamvec(2)
      endif
      name='dat.05'
      call histogram(-1d0,1d0,x,y,5,n,name,50,steps)
c decay angle distributions
      x=2d0
      if(kpvec*pvec(3).gt.0d0)then
        x=(kp(1)*p(3,1)+kp(2)*p(3,2)+kp(3)*p(3,3))/kpvec/pvec(3)
      endif
      name='dat.06'
      call histogram(-1d0,1d0,x,y,6,n,name,50,steps)
      x=2d0
      if(kpvec*pvec(4).gt.0d0)then
        x=(kp(1)*p(4,1)+kp(2)*p(4,2)+kp(3)*p(4,3))/kpvec/pvec(4)
      endif
      name='dat.07'
      call histogram(-1d0,1d0,x,y,7,n,name,50,steps)
      if(kmvec*pvec(5).gt.0d0)then
        x=(km(1)*p(5,1)+km(2)*p(5,2)+km(3)*p(5,3))/kmvec/pvec(5)
      endif
      name='dat.08'
      call histogram(-1d0,1d0,x,y,8,n,name,50,steps)
      x=2d0
      if(kmvec*pvec(6).gt.0d0)then
        x=(km(1)*p(6,1)+km(2)*p(6,2)+km(3)*p(6,3))/kmvec/pvec(6)
      endif
      name='dat.09'
      call histogram(-1d0,1d0,x,y,9,n,name,50,steps)
c weight for photon distributions
      yphoton=0d0
      if((smc.eq.2.and.nw.eq.6).or.(smc.ne.2.and.nw.eq.7))yphoton=y
c photon-energy distributions
      name='dat.10'
      call histogram(0d0,50d0,p(7,0),yphoton,10,n,name,50,steps)        
c photon-production angle
      if(pbeamvec(1)*pvec(7).gt.0d0)then   
        x=(pbeam(1,1)*p(7,1)+pbeam(1,2)*p(7,2)+pbeam(1,3)*p(7,3))
     *    /pbeamvec(1)/pvec(7)
      else
        x=1d0-0.04d0/10d0
      endif
      name='dat.11'
      call histogram(-1d0,1d0,x,yphoton,11,n,name,50,steps)
c minimal angle between photon and charged final-state particles
      if(pvec(7).gt.0d0)then
        x=200d0
        do i1=3,6
          if(q(i1).ne.0d0.and.pvec(i1).ne.0d0)then
            x=min(x,dacos((p(i1,1)*p(7,1)+p(i1,2)*p(7,2)
     *                    +p(i1,3)*p(7,3))/pvec(i1)/pvec(7))*180d0/pi)
          endif
        enddo
      else
        x=0d0  
      endif
      name='dat.12'
      call histogram(0d0,180d0,x,yphoton,12,n,name,50,steps)
c energy distributions of final-state particles
      name='dat.13'
      call histogram(0d0,energy*0.5d0,p(3,0),y,13,n,name,50,steps)     
      name='dat.14'
      call histogram(0d0,energy*0.5d0,p(4,0),y,14,n,name,50,steps)     
      name='dat.15'
      call histogram(0d0,energy*0.5d0,p(5,0),y,15,n,name,50,steps)     
      name='dat.16'
      call histogram(0d0,energy*0.5d0,p(6,0),y,16,n,name,50,steps) 
c decay angle in cmf of w boson
      x=2d0
      if(p(3,0).ne.0d0.and.kpvec.gt.0d0)then
        do i1=0,3
          q1(i1)=p(3,i1)
        enddo
        call boostrestframe(kp,q1)
        x=(q1(1)*kp(1)+q1(2)*kp(2)+q1(3)*kp(3))
     *    /dsqrt(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3))/kpvec
      endif
      name='dat.17'
      call histogram(-1d0,1d0,x,y,17,n,name,50,steps)
      x=2d0
      if(p(4,0).ne.0d0.and.kpvec.gt.0d0)then
        do i1=0,3
          q1(i1)=p(4,i1)
        enddo
        call boostrestframe(kp,q1)
        x=(q1(1)*kp(1)+q1(2)*kp(2)+q1(3)*kp(3))
     *    /dsqrt(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3))/kpvec
      endif
      name='dat.18'
      call histogram(-1d0,1d0,x,y,18,n,name,50,steps)
      x=2d0 
      if(p(5,0).ne.0d0.and.kmvec.gt.0d0)then
        do i1=0,3
          q1(i1)=p(5,i1)
        enddo
        call boostrestframe(km,q1)
        x=(q1(1)*km(1)+q1(2)*km(2)+q1(3)*km(3))
     *    /dsqrt(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3))/kmvec
      endif
      name='dat.19'
      call histogram(-1d0,1d0,x,y,19,n,name,50,steps) 
      x=2d0 
      if(p(6,0).ne.0d0.and.kmvec.gt.0d0)then
        do i1=0,3
          q1(i1)=p(6,i1)
        enddo
        call boostrestframe(km,q1)
        x=(q1(1)*km(1)+q1(2)*km(2)+q1(3)*km(3))
     *    /dsqrt(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3))/kmvec
      endif
      name='dat.20'
      call histogram(-1d0,1d0,x,y,20,n,name,50,steps) 
c angle between the decay planes
      q1(1)=kp(2)*p(3,3)-kp(3)*p(3,2)
      q1(2)=kp(3)*p(3,1)-kp(1)*p(3,3)
      q1(3)=kp(1)*p(3,2)-kp(2)*p(3,1)
      q2(1)=-km(2)*p(5,3)+km(3)*p(5,2)
      q2(2)=-km(3)*p(5,1)+km(1)*p(5,3)
      q2(3)=-km(1)*p(5,2)+km(2)*p(5,1)
      denum=dsqrt(dabs(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3)))
     *     *dsqrt(dabs(q2(1)*q2(1)+q2(2)*q2(2)+q2(3)*q2(3)))
      x=-1d0
      if(denum.ne.0d0)then
        x=(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/denum
        if(dabs(x).le.1d0)then
          x=dacos(x)*180d0/pi
          q3(1)=q1(2)*q2(3)-q1(3)*q2(2)
          q3(2)=q1(3)*q2(1)-q1(1)*q2(3)
          q3(3)=q1(1)*q2(2)-q1(2)*q2(1)
          if((kp(1)*q3(1)+kp(2)*q3(2)+kp(3)*q3(3)).lt.0d0)x=360d0-x
        else
          x=-1d0
        endif
      endif
      name='dat.21'
      call histogram(0d0,360d0,x,y,21,n,name,50,steps)
c polar angle of w decay 
      q1(1)=kp(2)*pbeam(1,3)-kp(3)*pbeam(1,2)
      q1(2)=kp(3)*pbeam(1,1)-kp(1)*pbeam(1,3)
      q1(3)=kp(1)*pbeam(1,2)-kp(2)*pbeam(1,1)
      q2(1)=kp(2)*p(3,3)-kp(3)*p(3,2)
      q2(2)=kp(3)*p(3,1)-kp(1)*p(3,3)
      q2(3)=kp(1)*p(3,2)-kp(2)*p(3,1)
      denum=dsqrt(dabs(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3)))
     *     *dsqrt(dabs(q2(1)*q2(1)+q2(2)*q2(2)+q2(3)*q2(3)))
      x=-1d0
      if(denum.ne.0d0)then
        x=(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/denum
        if(dabs(x).le.1d0)then
          x=dacos(x)*180d0/pi
          q3(1)=q1(2)*q2(3)-q1(3)*q2(2)
          q3(2)=q1(3)*q2(1)-q1(1)*q2(3)
          q3(3)=q1(1)*q2(2)-q1(2)*q2(1)
          if((kp(1)*q3(1)+kp(2)*q3(2)+kp(3)*q3(3)).lt.0d0)x=360d0-x
        else
          x=-1d0
        endif
      endif
      name='dat.22'
      call histogram(0d0,360d0,x,y,22,n,name,50,steps)
      q1(1)=km(2)*pbeam(2,3)-km(3)*pbeam(2,2)
      q1(2)=km(3)*pbeam(2,1)-km(1)*pbeam(2,3)
      q1(3)=km(1)*pbeam(2,2)-km(2)*pbeam(2,1)
      q2(1)=km(2)*p(5,3)-km(3)*p(5,2)
      q2(2)=km(3)*p(5,1)-km(1)*p(5,3)
      q2(3)=km(1)*p(5,2)-km(2)*p(5,1)
      denum=dsqrt(dabs(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3)))
     *     *dsqrt(dabs(q2(1)*q2(1)+q2(2)*q2(2)+q2(3)*q2(3)))
      x=-1d0
      if(denum.ne.0d0)then
        x=(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/denum
        if(dabs(x).le.1d0)then
          x=dacos(x)*180d0/pi
          q3(1)=q1(2)*q2(3)-q1(3)*q2(2)
          q3(2)=q1(3)*q2(1)-q1(1)*q2(3)
          q3(3)=q1(1)*q2(2)-q1(2)*q2(1)
          if((km(1)*q3(1)+km(2)*q3(2)+km(3)*q3(3)).lt.0d0)x=360d0-x
        else
          x=-1d0
        endif
      endif
      name='dat.23'
      call histogram(0d0,360d0,x,y,23,n,name,50,steps)
c angle between the fermions corresponding to one W boson
      x=2d0
      if(pvec(3)*pvec(4).gt.0d0)then
        x=(p(3,1)*p(4,1)+p(3,2)*p(4,2)+p(3,3)*p(4,3))/pvec(3)/pvec(4)
      endif
      name='dat.24'
      call histogram(-1d0,1d0,x,y,24,n,name,50,steps) 
      x=2d0
      if(pvec(5)*pvec(6).gt.0d0)then
        x=(p(5,1)*p(6,1)+p(5,2)*p(6,2)+p(5,3)*p(6,3))/pvec(5)/pvec(6)
      endif
      name='dat.25'
      call histogram(-1d0,1d0,x,y,25,n,name,50,steps) 
c W+ production angle in the WW CMF:
      x=2d0
      do i1=0,3
         q1(i1)=pbeam(1,i1)
         q2(i1)=kp(i1)
         q3(i1)=kp(i1)+km(i1)
      enddo
      call boostrestframe(q3,q1)
      call boostrestframe(q3,q2)
      q1vec=dsqrt(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3))
      q2vec=dsqrt(q2(1)*q2(1)+q2(2)*q2(2)+q2(3)*q2(3))
      if(q1vec*q2vec.gt.0d0)then
        x=(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/q1vec/q2vec
      endif
      name='dat.26'
      call histogram(-1d0,1d0,x,y,26,n,name,50,steps)
c W- production angle in the WW CMF:
      x=2d0
      do i1=0,3
         q1(i1)=pbeam(2,i1)
         q2(i1)=km(i1)
         q3(i1)=kp(i1)+km(i1)
      enddo
      call boostrestframe(q3,q1)
      call boostrestframe(q3,q2)
      q1vec=dsqrt(q1(1)*q1(1)+q1(2)*q1(2)+q1(3)*q1(3))
      q2vec=dsqrt(q2(1)*q2(1)+q2(2)*q2(2)+q2(3)*q2(3))
      if(q1vec*q2vec.gt.0d0)then
        x=(q1(1)*q2(1)+q1(2)*q2(2)+q1(3)*q2(3))/q1vec/q2vec
      endif
      name='dat.27'
      call histogram(-1d0,1d0,x,y,27,n,name,50,steps)
c your own histograms can be added here, e.g.:
c x=observable
c name='dat.28'
c call histogram(min.value of x,max.value of x,x,y,28,n,name,50,steps)
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     histogram routine                                           c
c     Markus Roth, 11. August 2000                                c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine histogram(min,max,x,y,histo,n,name,parts,steps)
      implicit none
c local variable
      integer maxh,maxp
      parameter(maxh=50,maxp=1000)
      real*8 average(maxh,maxp),sigma(maxh,maxp),ysum(maxh,maxp)
      real*8 avborn(maxh,maxp),siborn(maxh,maxp)
      real*8 min,max,x,y,av,si,avb,sib,shift
      integer i,isum(maxh,maxp),imax(maxh)
      integer i1,i2,histo,parts,n,nout,steps
      character*10 name
      save average,sigma,ysum,isum,imax
c initialization
      if(steps.eq.1)then
        do i1=1,maxh
          imax(i1)=0
          do i2=1,maxp
            ysum(i1,i2)=0d0
            average(i1,i2)=0d0
            sigma(i1,i2)=0d0
            avborn(i1,i2)=0d0
            siborn(i1,i2)=0d0
            isum(i1,i2)=0
          enddo
        enddo
      endif
c input of single subevents
      if(steps.eq.2)then
        if(x.ge.min.and.x.lt.max)then
          i=int((x-min)/(max-min)*parts)+1
          ysum(histo,i)=y+ysum(histo,i)
          imax(histo)=imax(histo)+1
          isum(histo,imax(histo))=i
        endif
c input born event and final calculation of whole event
      elseif(steps.eq.3)then        
        if(x.ge.min.and.x.lt.max)then
          i=int((x-min)/(max-min)*parts)+1
          avborn(histo,i)=y+avborn(histo,i)
          siborn(histo,i)=y*y+siborn(histo,i)
        endif
        if(imax(histo).ge.1)then
          do i1=1,imax(histo)
            i=isum(histo,i1)
            average(histo,i)=ysum(histo,i)+average(histo,i)
            sigma(histo,i)=ysum(histo,i)*ysum(histo,i)+sigma(histo,i)
            ysum(histo,i)=0d0
          enddo
        endif
        imax(histo)=0
c output
      elseif(steps.eq.4)then
        nout=25
        open(unit=nout,file=name,status='unknown')
        shift=1d0
        write(nout,'("# histogram:",i3,", bins:",i4)'),histo,parts
        write(nout,'(a)')'# '
        do i=1,parts
          av=average(histo,i)/dble(n)/(max-min)*dble(parts)
          si=sigma(histo,i)/(dble(n)*(max-min)/dble(parts))**2
          si=dsqrt(dabs(si-av*av/dble(n)))
          avb=avborn(histo,i)/dble(n)/(max-min)*dble(parts)
          sib=siborn(histo,i)/(dble(n)*(max-min)/dble(parts))**2
          sib=dsqrt(dabs(sib-avb*avb/dble(n)))
          write(nout,'(5(e12.6,2x))') 
     *      min+(max-min)/dble(parts)*(dble(i)-shift),av,si,avb,sib
        enddo
        write(nout,'(a)') '# '
        if(shift.eq.0d0)write(nout,'(a)')'# x = right corner'
        if(shift.eq.0.5d0)write(nout,'(a)')'# x = mid'
        if(shift.eq.1d0)write(nout,'(a)')'# x = left corner'
        close(nout)
      endif
      end

ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                 c
c     auxiliary subroutine for QGC numerics                       c
c     calculation of neutrino momentum                            c
c     invariant mass cut: Mff' = MW +/- 2 GW                      c
c                                                                 c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine qgc(pbeam,p,y,nneutrino)
      implicit none
c local variable
      real*8 pbeam(2,0:3),p(7,0:3),y,kp(0:3),km(0:3),kp2,km2
      integer nneutrino,i1
c param
      real*8 pi,el,alpha,alpha0,alphaz,alphas,gf,cw,cw2,sw,sw2,mw,mw2
      real*8 gw,mz,mz2,gz,mh,mh2,mxe,mxe2,ml(3),ml2(3),mqp(3)
      real*8 mqp2(3),mqm(3),mqm2(3)
      common/param/pi,el,alpha,alpha0,alphaz,alphas,gf,cw,cw2,sw,sw2,
     *  mw,mw2,gw,mz,mz2,gz,mh,mh2,mxe,mxe2,ml,ml2,mqp,mqp2,mqm,mqm2
      if (y.eq.0d0) return 
      do i1=0,3
        p(nneutrino,i1)=0d0 
        p(nneutrino,i1)=pbeam(1,i1)+pbeam(2,i1)
     *    -p(3,i1)-p(4,i1)-p(5,i1)-p(6,i1)-p(7,i1)
        kp(i1)=p(3,i1)+p(4,i1)
        km(i1)=p(5,i1)+p(6,i1)
      enddo
      kp2=kp(0)**2-kp(1)**2-kp(2)**2-kp(3)**2
      km2=km(0)**2-km(1)**2-km(2)**2-km(3)**2
      if(dabs(dsqrt(dabs(kp2))-mw).gt.2d0*gw)y=0d0
      if(dabs(dsqrt(dabs(km2))-mw).gt.2d0*gw)y=0d0
      end

cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c                                                                c
c     boost particle with momentum p into restframe              c
c     of particle with momentum q                                c
c                                                                c
cccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
      subroutine boostrestframe(q,p)
      implicit none
c local variable
      real*8 p(0:3),q(0:3),m,m2,bx,by,bz,gamma
      real*8 p0,px,py,pz,bp,a
      m2=q(0)*q(0)-q(1)*q(1)-q(2)*q(2)-q(3)*q(3)
      if(m2.le.0d0)return
      m=dsqrt(dabs(m2))
      bx=-q(1)/m
      by=-q(2)/m
      bz=-q(3)/m
      gamma=q(0)/m
      a=1d0/(1d0+gamma)
      p0=p(0)
      px=p(1)
      py=p(2)
      pz=p(3)
      bp=bx*px+by*py+bz*pz
      p(0)=gamma*p0+bp
      p(1)=px+bx*p0+a*bp*bx
      p(2)=py+by*p0+a*bp*by
      p(3)=pz+bz*p0+a*bp*bz
      end
